/*-
 * Copyright (C) @BABOLO  2002 Dec 17
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char copyright[] = "\
@(#)Copyright (C) @BABOLO  2002 Dec 17\n\
@(#)All rights reserved.\n";
static const char rcsid[] = "$Id: cmd_select.c,v 1.10 2007/11/10 22:00:36 babolo Exp $";
#endif /* not lint */

#include <sys/types.h>
#include <unistd.h>
#include <stdlib.h>
#include <limits.h>
#include <string.h>
#include <stdio.h>
#include <mife.h>
#include <err.h>
#include <libpq-fe.h>
#include <babolo/BLINflag.h>
#include "pgoblin.h"
#include "pgob.h"

#define cmd (options->pgm->code[options->ppoint].cmd)
typedef enum
{   tbeg
,   tend
,   rbeg
,   rmed
,   rend
,   cbeg
,   cmed
,   cend
} st;

int
pgoblin_select(pgoblin_con *options, int a[]) {
    char *q;

    ifBLIN_QV3(options->flags) fprintf( stderr, "+pgoblin_select %d %d %d =%s~\n"
                                      , a[PGO_POUT], a[PGO_PCTL], a[PGO_PCON]
                                      , (char*)IO_CTL.text
                                      );
    q = IO_CTL.text;
    if  (!q) {
        ifBLIN_QV1(options->flags) warnx("pgoblin_select: No query");
        return(-1);
    }
    PQEXEC(IO_OUT.pq, q, PGRES_TUPLES_OK, NULL);
    ifBLIN_QV3(options->flags)
        fprintf(stderr, "-pgoblin_select %s\n", IO_OUT.pq ? "OK" : "NULL");
    return(0);
}

static int
pm(pgoblin_con *options, int reg_out, int reg_style, int q, st p) {
    char *c = NULL;

    if  (options->style[reg_style]) {
        switch (p) {
        case tbeg: c = options->style[reg_style]->tbeg; break;
        case tend: c = options->style[reg_style]->tend; break;
        case rbeg: c = options->style[reg_style]->rbeg; break;
        case rmed: c = options->style[reg_style]->rmed; break;
        case rend: c = options->style[reg_style]->rend; break;
        case cbeg: c = options->style[reg_style]->cbeg; break;
        case cmed: c = options->style[reg_style]->cmed; break;
        case cend: c = options->style[reg_style]->cend; break;
        }
        if  (c && q) return(options->io[reg_out].wri(options->io[reg_out].onu, c, 0));
          else return(0);
    } else return(0);
}

int
pgoblin_display(pgoblin_con *options, int a[]) {
    int j, c, k;
    ssize_t e;

    ifBLIN_QV3(options->flags)
        fprintf(stderr, "+pgoblin_display %d %d %d\n", a[PGO_POUT], a[PGO_PSTY], a[PGO_PIN]);
    j = e = 0;
    if  (e >= 0) e = pm(options, a[PGO_POUT], a[PGO_PSTY], 1, tbeg);
    for (k = 0; k < PQntuples(IO_IN.pq) && e >= 0; k++) {
        if  (e >= 0) e = pm(options, a[PGO_POUT], a[PGO_PSTY], k, rmed);
        if  (e >= 0) e = pm(options, a[PGO_POUT], a[PGO_PSTY], 1, rbeg);
        for (c = 0; c < PQnfields(IO_IN.pq) && e >= 0; c++) {
            if  (e >= 0) e = pm(options, a[PGO_POUT], a[PGO_PSTY], c, cmed);
            if  (e >= 0) e = pm(options, a[PGO_POUT], a[PGO_PSTY], 1, cbeg);
            if  (e >= 0 && !PQgetisnull(IO_IN.pq, k, c))
                j += (e = IO_OUT.wri( IO_OUT.onu
                                    , PQgetvalue(IO_IN.pq, k, c)
                                    , PQgetlength(IO_IN.pq, k, c)
                     )              );
            if  (e >= 0) e = pm(options, a[PGO_POUT], a[PGO_PSTY], 1, cend);
        }
        if  (e >= 0) e = pm(options, a[PGO_POUT], a[PGO_PSTY], 1, rend);
    }
    if  (e >= 0) e = pm(options, a[PGO_POUT], a[PGO_PSTY], 1, tend);
    if  (e < 0) {
        ifBLIN_QV1(options->flags) warnx("pgoblin_display: write error");
        return(-1);
    }
    ifBLIN_QV3(options->flags) fprintf(stderr, "-pgoblin_display %d\n", 0);
    return(0);
}

static const char *fetchN[] =
{ "FETCH 1 FROM pgoblin"
, "FETCH 2 FROM pgoblin"
, "FETCH 3 FROM pgoblin"
, "FETCH 5 FROM pgoblin"
, "FETCH 8 FROM pgoblin"
, "FETCH 13 FROM pgoblin"
, "FETCH 21 FROM pgoblin"
, "FETCH 34 FROM pgoblin"
, "FETCH 55 FROM pgoblin"
, "FETCH 89 FROM pgoblin"
, "FETCH 144 FROM pgoblin"
, "FETCH 233 FROM pgoblin"
, "FETCH 377 FROM pgoblin"
, "FETCH 610 FROM pgoblin"
, "FETCH 987 FROM pgoblin"
, "FETCH 1597 FROM pgoblin"
, "FETCH 2584 FROM pgoblin"
, "FETCH 4181 FROM pgoblin"
, "FETCH 6765 FROM pgoblin"
, "FETCH 10946 FROM pgoblin"
, "FETCH 17711 FROM pgoblin"
, NULL
};

int
pgoblin_strsel(pgoblin_con *options, int a[]) {
    int j, c, k, t;
    PGresult *res;
    char *q;
    size_t i;
    ssize_t e;

    ifBLIN_QV3(options->flags) fprintf( stderr, "+pgoblin_strsel %d %d %d %d =%s~\n"
                                      , a[PGO_POUT], a[PGO_PCTL], a[PGO_PCON], a[PGO_PSTY]
                                      , (char*)IO_CTL.text
                                      );
    if  (!(q = IO_CTL.text)) {
        ifBLIN_QV1(options->flags) warnx("pgoblin_strsel: No query");
        return(-1);
    }

    if  (!R_CONN.intran) {
        PQEXEC(res, "BEGIN", PGRES_COMMAND_OK, -1);
        PQclear(res);
    }
    asprintf(&q, "DECLARE pgoblin NO SCROLL CURSOR FOR %s", q);
    if  (q == NULL) {
        ifBLIN_QV1(options->flags) warnx("pgoblin_strsel: No memory");
        return(-1);
    }
    PQEXEC(res, q, PGRES_COMMAND_OK, -1);
    PQclear(res);
    free(q);
    i = 0;
    PQEXEC(res, fetchN[i], PGRES_TUPLES_OK, -1);
    j = e = 0;
    if  (e >= 0) e = pm(options, a[PGO_POUT], a[PGO_PSTY], 1, tbeg);

    for (j = 0; (t = PQntuples(res)) > 0 && e >= 0; j = 0) {
        for (k = 0; k < t && e >= 0; k++) {
            if  (e >= 0) e = pm(options, a[PGO_POUT], a[PGO_PSTY], (k && i), rmed);
            if  (e >= 0) e = pm(options, a[PGO_POUT], a[PGO_PSTY], 1, rbeg);
            for (c = 0; c < PQnfields(res) && e >= 0; c++) {
                if  (e >= 0) e = pm(options, a[PGO_POUT], a[PGO_PSTY], c, cmed);
                if  (e >= 0) e = pm(options, a[PGO_POUT], a[PGO_PSTY], 1, cbeg);
                if  (e >= 0 && !PQgetisnull(res, k, c))
                    j += (e = IO_OUT.wri( IO_OUT.onu
                                        , PQgetvalue(res, k, c)
                                        , PQgetlength(res, k, c)
                         )              );
                if  (e >= 0) e = pm(options, a[PGO_POUT], a[PGO_PSTY], 1, cend);
            }
            if  (e >= 0) e = pm(options, a[PGO_POUT], a[PGO_PSTY], 1, rend);
        }
        if  (e >= 0) e = pm(options, a[PGO_POUT], a[PGO_PSTY], 1, tend);
        PQclear(res);
        ifBLIN_QV3(options->flags)
            fprintf(stderr, "j=%d aim=%d i %d => ", j, PGOBLIN_BUFMULT * PGOBLIN_BUFSIZE, i);
        if  (j > PGOBLIN_BUFMULT * PGOBLIN_BUFSIZE && i > 1) i--;
        if  (j < PGOBLIN_BUFMULT * PGOBLIN_BUFSIZE && fetchN[i + 1]) i++;
        ifBLIN_QV3(options->flags) fprintf(stderr, "%d\n", i);
        PQEXEC(res, fetchN[i], PGRES_TUPLES_OK, -1)
    }

    PQclear(res);
    PQEXEC(res, "CLOSE pgoblin", PGRES_COMMAND_OK, -1);
    PQclear(res);
    if  (!R_CONN.intran) {
        PQEXEC(res, "END", PGRES_COMMAND_OK, -1);
        PQclear(res);
    }
    if  (e < 0) {
        ifBLIN_QV1(options->flags) warnx("pgoblin_strsel: write error");
        return(-1);
    }
    ifBLIN_QV3(options->flags) fprintf( stderr, "-pgoblin_strsel %d\n", 0);
    return(0);
}

int
pgoblin_binsel(pgoblin_con *options, int a[]) {
    int j, c, k, t;
    PGresult *res;
    char *q;
    size_t i;
    ssize_t e;

    ifBLIN_QV3(options->flags) fprintf( stderr, "+pgoblin_binsel %d %d %d %d =%s~\n"
                                      , a[PGO_POUT], a[PGO_PCTL], a[PGO_PCON], a[PGO_PSTY]
                                      , (char*)IO_CTL.text
                                      );
    if  (!(q = IO_CTL.text)) {
        ifBLIN_QV1(options->flags) warnx("pgoblin_binsel: No query");
        return(-1);
    }

    if  (!R_CONN.intran) {
/*  if  (PQtransactionStatus(R_CONN.conn) != PQTRANS_INTRANS) { */
        PQEXEC(res, "BEGIN", PGRES_COMMAND_OK, -1);
        PQclear(res);
    }
    asprintf(&q, "DECLARE pgoblin BINARY NO SCROLL CURSOR FOR %s", q);
    if  (q == NULL) {
        ifBLIN_QV1(options->flags) warnx("pgoblin_binsel: No memory");
        return(-1);
    }
    PQEXEC(res, q, PGRES_COMMAND_OK, -1);
    PQclear(res);
    free(q);
    i = 0;
    PQEXEC(res, fetchN[i], PGRES_TUPLES_OK, -1);

    for (e = 0; (t = PQntuples(res)) > 0 && e >= 0; ) {
        for (j = 0, k = 0; k < t && e >= 0; k++) {
            for (c = 0; c < PQnfields(res) && e >= 0; c++) {
                if  (e >= 0 && !PQgetisnull(res, k, c))
                    j += (e = IO_OUT.wri( IO_OUT.onu
                                        , PQgetvalue(res, k, c)
                                        , PQgetlength(res, k, c)
                         )              );
        }   }
        PQclear(res);
        ifBLIN_QV3(options->flags)
            fprintf(stderr, "j=%d aim=%d i %d => ", j, PGOBLIN_BUFMULT * PGOBLIN_BUFSIZE, i);
        if  (j > PGOBLIN_BUFMULT * PGOBLIN_BUFSIZE && i > 0) i--;
        if  (j < PGOBLIN_BUFMULT * PGOBLIN_BUFSIZE && fetchN[i + 1]) i++;
        ifBLIN_QV3(options->flags) fprintf(stderr, "%d\n", i);
        PQEXEC(res, fetchN[i], PGRES_TUPLES_OK, -1)
    }

    PQclear(res);
    PQEXEC(res, "CLOSE pgoblin", PGRES_COMMAND_OK, -1);
    PQclear(res);
    if  (!R_CONN.intran) {
        PQEXEC(res, "END", PGRES_COMMAND_OK, -1);
        PQclear(res);
    }
    if  (e < 0) {
        ifBLIN_QV1(options->flags) warnx("pgoblin_binsel: write error");
        return(-1);
    }
    ifBLIN_QV3(options->flags) fprintf( stderr, "-pgoblin_binsel %d\n", 0);
    return(0);
}

int
pgoblin_binpre(pgoblin_con *options, int a[]) {
    PGresult *res;
    char *q;

    ifBLIN_QV3(options->flags) fprintf( stderr, "+pgoblin_binpre %d %d %d =%s~\n"
                                      , a[PGO_POUT], a[PGO_PCTL], a[PGO_PCON]
                                      , (char*)IO_CTL.text
                                      );
    if  (!(q = IO_CTL.text)) {
        ifBLIN_QV1(options->flags) warnx("pgoblin_binpre: No query");
        return(-1);
    }

    if  (!R_CONN.intran) {
/*  if  (PQtransactionStatus(R_CONN.conn) != PQTRANS_INTRANS) { */
        PQEXEC(res, "BEGIN", PGRES_COMMAND_OK, -1);
        PQclear(res);
    }
    asprintf(&q, "DECLARE pgoblin BINARY NO SCROLL CURSOR FOR %s", q);
    if  (q == NULL) {
        ifBLIN_QV1(options->flags) warnx("pgoblin_binpre: No memory");
        return(-1);
    }
    PQEXEC(res, q, PGRES_COMMAND_OK, -1);
    PQclear(res);
    free(q);
    PQEXEC(IO_OUT.pq, "FETCH 2147483647 FROM pgoblin", PGRES_TUPLES_OK, -1);
    PQEXEC(res, "CLOSE pgoblin", PGRES_COMMAND_OK, -1);
    PQclear(res);
    if  (!R_CONN.intran) {
        PQEXEC(res, "END", PGRES_COMMAND_OK, -1);
        PQclear(res);
    }
    ifBLIN_QV3(options->flags) fprintf( stderr, "-pgoblin_binpre %d\n", 0);
    return(0);
}
