/*-
 * Copyright (C) @BABOLO  2002 Dec 23
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char copyright[] = "\
@(#)Copyright (C) @BABOLO  2002 Dec 23\n\
@(#)All rights reserved.\n";
static const char rcsid[] = "$Id: parser.c,v 1.14 2007/11/09 07:23:09 babolo Exp $";
#endif /* not lint */

#include <sys/types.h>
#include <sysexits.h>
#include <unistd.h>
#include <stdlib.h>
#include <limits.h>
#include <string.h>
#include <stdio.h>
#include <errno.h>
#include <mife.h>
#include <err.h>
#include <libpq-fe.h>
#include <babolo/parser.h>
#include <babolo/BLINflag.h>
#include "pgoblin.h"
#include "pgob.h"
#include "cmds.h"

#define L0 0 /* \0           */
#define Ll 1 /*   */
#define Lb 2 /*        */
#define Lp 3 /*    */
#define Ld 4 /* #            */
#define Lo 5 /*     */
#define Lx 6
static const u_char *clasnames[] =
{ "L0", "Ll", "Lb", "Lp", "Ld", "Lo", "Lx"};

static const char class[256] =
{ L0, Lo, Lo, Lo,  Lo, Lo, Lo, Lo,  Lo, Lb, Ll, Ll,  Lo, Ll, Lo, Lo
, Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo
, Lb, Lo, Lo, Ld,  Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo
, -1, -2, -3, -4,  -5, -6, -7, -8,  -9,-10, Lo, Lo,  Lo, Lo, Lo, Lo

, Lo,-11,-12,-13, -14,-15,-16,-17, -18,-19,-20,-21, -22,-23,-24,-25
,-26,-27,-28,-29, -30,-31,-32,-33, -34,-35,-36, Lo,  Lo, Lo, Lo,-63
, Lo,-37,-38,-39, -40,-41,-42,-43, -44,-45,-46,-47, -48,-49,-50,-51
,-52,-53,-54,-55, -56,-57,-58,-59, -60,-61,-62, Lo,  Lo, Lo, Lo, Lo

, Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo
, Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo
, Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo
, Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo

, Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo
, Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo
, Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo
, Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo
};

static const u_char *statenames[] =
{ "bg", "bs", "nl", "tc", "tb", "tp", "sc", "ml", "mc", "xx"};

enum states
{ bg /*                                */
, bs /*                                 */
, nl /*                           */
, tc /*                 */
, tb /*         */
, tp /*                                */
, sc /*       () */
, ml /*                           */
, mc /*                                 */
, xx
};
static const u_char *pn[] = 
{ "- ", "- ", "- ", "- ", "- ", "- ", "- ", "- "
, "Cl", "Cc", "Cp", "- ", "Cb", "Ce", "- ", "Ci"
, "Co", "- ", "- ", "- ", "Cx", "Cy", "- ", "- "
};

#define Cl  0x800000 /*  lexor */
#define Cp  0x200000 /*   */
#define Cb  0x080000 /*    */
#define Ce  0x040000 /*    */
#define Ci  0x010000 /*     */
#define Co  0x008000 /*   */
#define Cx  0x000800 /*   */
#define Cy  0x000400 /*  */
#define C_state 0xFF /*   */

static u_int32_t automa[Lx][xx] =
/*   bg    bs    nl       tc    tb    tp    sc       ml    mc  */
{{Cx|xx,Cx|xx,Cx|xx,Cl|Ci|xx,Ci|xx,Ci|xx,Ci|xx,Ci|   xx,Cy|xx}/* L0 \0           */
,{   bs,   nl,   nl,Cl|   ml,   ml,   ml,   ml,Ce|Cb|ml,Ce|ml}/* Ll   */
,{   bs,   bs,   bs,Cl|   tb,   tb,   sc,   sc,   Cb|mc,   mc}/* Lb        */
,{   bs,   bs,   bs,Cl|   tc,Cp|tp,Cp|tp,   sc,   Cb|mc,   mc}/* Lp    */
,{Co|tc,   bs,Co|tc,Cl|   sc,   sc,   sc,   sc,Ci|Co|tc,   mc}/* Ld #            */
,{   bs,   bs,   bs,Cl|   sc,   sc,   sc,   sc,   Cb|mc,   mc}/* Lo     */
};

static const u_char paramparam[/* PGO_C* */]={PGO_POUT, PGO_PCTL, PGO_PIN, PGO_PCON};

int
pgoblin_parser(pgoblin_con *options, pgoblin_p *pgm, mife_descriptor *in) {
    int ex = EX_OK;
    off_t offset;
    enum states state;
    u_int32_t control = 0;
    const u_char **qq, *qp;
    u_char *q = NULL, s, *r_beg = NULL, *r_end = NULL;
    char p;
    u_int32_t r_lex = 0, r_cmd = 0, r_parm = 0;
    int r_pp = 0, lfcount, i;

    ifBLIN_QV3(options->flags) fprintf( stderr
                                      , "+pgoblin_parser: options %s, pgm %s, in %s\n"
                                      , options ? "OK" : "NULL"
                                      , pgm ? "OK" : "NULL"
                                      , in ? "OK" : "NULL"
                                      );
    ifBLIN_QV4(options->flags) fprintf(stderr, "pgm@%d:%d\n", pgm->curr, pgm->maxx);
    ifBLIN_QV5(options->flags) BLIN_STATEHEADER(stderr, pn)
    lfcount = 0;
    qp = &s;
    qq = &qp;
    for (state = bg, offset = 0; state < xx; offset++, state = control & C_state) {
        if  ((mife_read(in, 0, offset) > 0) && (q = mife_get(in, offset))) {
            s = *q;
            p = class[s];
            if  (p < 0) p = Lp;
            if  (s == '\n') lfcount++;
        } else {
            p = 0;
            s = 0;
        }
        control = automa[(u_char)p][state];
        ifBLIN_QV5(options->flags)
            BLIN_STATEBODY(stderr, pn, r_parm & 0xFF, clasnames[(u_char)p], q, 0, r_lex, control, statenames, state, C_state, *(q + BLIN_STATEVAR_i)>=' ')
        if  (control & Cl) {
            qp = &s;
            r_lex = babolo_testchar(&cmds, r_lex, qq);
        }
        if  (control & Cp) r_parm |= (~class[s]) << (r_pp += 6);
        if  ((control & Cb) && !r_beg) r_beg = q;
        if  (control & Ce && r_beg) r_end = q;
        if  (control & Ci) {
            if  (r_lex >= cmds.szt) r_cmd = ~r_lex;
            ifBLIN_QV3(options->flags) fprintf(stderr, "[%d]=%02X(%08X)\n", pgm->curr, r_cmd, r_parm);
            if  (pgoblin_a[r_cmd].flags & PGOBLIN_ILLEGAL) {
                errx(EX_SOFTWARE, "Illegal command at line %d", lfcount);
            } else if (pgm->curr > pgm->maxx) {
                errx(EX_SOFTWARE, "Programm too big");
            } else if (r_cmd > PGOBLIN_max) {
                errx(EX_SOFTWARE, "Illegal command at line %d", lfcount);
            } else if (!(pgoblin_a[r_cmd].flags & PGOBLIN_IGNORE)) {
                u_int32_t cc;

                cc = 0;
                for (i = 0; i < 4; i++) {
                    if  (pgoblin_a[r_cmd].flags & (PGOBLIN_ISOUT << i))
                        cc |= ((r_parm & 0x3F) << (paramparam[i] * 6));
                    r_parm >>= 6;
                }
                if  (pgoblin_a[r_cmd].flags & PGOBLIN_ISSTYLE)
                    cc |= ((r_parm & 0x3F) << (pgoblin_a[r_cmd].flags & PGOBLIN_STYLEN));
                pgm->code[pgm->curr].cmd = r_cmd | (cc << 8);
                pgm->code[pgm->curr].lit = NULL;
                if  (r_beg && r_end) {
                    *r_end = '\0';
                    if  (!(pgoblin_a[r_cmd].flags & PGOBLIN_NOLITER)) {
                        pgm->code[pgm->curr].lit = r_beg;
                    }
                    ifBLIN_QV3(options->flags) {
                        if  (pgm->code[pgm->curr].lit)
                            fprintf(stderr, ">%s~\n", pgm->code[pgm->curr].lit);
                }   }
                pgm->curr++;
        }   }
        if  (control & Co) {
            r_lex = 0;
            r_cmd = 0;
            r_parm = 0;
            r_pp = -6;
            r_beg = 0;
            r_end = 0;
        }
        if  (control & Cx) {
            errno = EINVAL;
            ex = (EX_DATAERR);
        }
        if  (control & Cy) {
            errno = EINVAL;
            ex = (EX_DATAERR);
    }   }
    ifBLIN_QV3(options->flags) fprintf(stderr, "-pgoblin_parser %d\n", ex);
    return(ex);
}
