/*-
 * Copyright (C) @BABOLO  2002 Dec 04
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: pgoblin.h,v 1.20 2007/11/10 22:00:36 babolo Exp $
 */

#define PGOBLIN_REGSIZE   64
#define PGOBLIN_PROGSIZE  8191

typedef enum
{ PGOBLIN_null
, PGOBLIN_cont
, PGOBLIN_bang
, PGOBLIN_connect
, PGOBLIN_listen
, PGOBLIN_begin
, PGOBLIN_end
, PGOBLIN_perform
, PGOBLIN_echo
, PGOBLIN_cat
, PGOBLIN_exec
, PGOBLIN_fork
, PGOBLIN_include
, PGOBLIN_copyin
, PGOBLIN_copyout
, PGOBLIN_binselect
, PGOBLIN_strselect
, PGOBLIN_binpresel
, PGOBLIN_select
, PGOBLIN_display
, PGOBLIN_getarg
, PGOBLIN_getenv
, PGOBLIN_getget
, PGOBLIN_getpost
, PGOBLIN_getproc
, PGOBLIN_max
} pgoblin_command;

typedef u_int32_t pgoblin_cmd;
#define PGOBLIN_COMMAND 0x000000FF
#define PGOBLIN_ARG1    0x00003F00
#define PGOBLIN_ARG2    0x000FC000
#define PGOBLIN_ARG3    0x03F00000
#define PGOBLIN_ARG4    0xFC000000
#define PGOBLIN_ARGMASK 0x0000003F
#define PGOBLIN_ARG1S   8
#define PGOBLIN_ARG2S   14
#define PGOBLIN_ARG3S   20
#define PGOBLIN_ARG4S   26

typedef struct pgoblin_t {
    pgoblin_cmd cmd;
    char       *lit;
} pgoblin_t;

typedef struct pgoblin_p {
    u_int32_t flags;
    ssize_t   maxx;
    ssize_t   curr;
    pgoblin_t code[1];
} pgoblin_p;

typedef struct pgoblin_io {
    u_int32_t        flags;
/* For pgoblin_c.ch[] too */
#   define           PGOBLIN_EMPTY    0x000000
#   define           PGOBLIN_BINPARM  0x000001
#   define           PGOBLIN_TEXTPARM 0x000002
#   define           PGOBLIN_FILENAME 0x000004
#   define           PGOBLIN_ANYTEXT  (PGOBLIN_BINPARM | PGOBLIN_TEXTPARM | PGOBLIN_FILENAME)
#   define           PGOBLIN_MIFEDES  0x000008
#   define           PGOBLIN_PQRESULT 0x000010
#   define           PGOBLIN_OUTSET   0x000020
#   define           PGOBLIN_STYLE    0x000040
#   define           PGOBLIN_CONNECT  0x000080
/* Byte ends */
#   define           PGOBLIN_FREETEXT 0x000200
#   define           PGOBLIN_FREEMIFE 0x000800
#   define           PGOBLIN_FREEPQRE 0x001000
#   define           PGOBLIN_FREEOUTS 0x002000
#   define           PGOBLIN_REFERER  0x010000
#   define           PGOBLIN_CTLARRAY 0x020000
/* For pgoblin_io_c(,,) */
#   define           PGOBLIN_FORCE0   0x400000
    ssize_t          ref;
    void            *text;
    size_t           length;
    size_t           offset;
    pid_t            pid;
    mife_descriptor *mife;
    PGresult        *pq;
    int            (*wri)(int onu, const void *buf, size_t size);
    int            (*clo)(int onu);
    int              onu;
} pgoblin_io;

/* Do not use style! Any style must be NULL */
typedef struct pgoblin_style {
    char *tbeg;
    char *tend;
    char *rbeg;
    char *rmed;
    char *rend;
    char *cbeg;
    char *cmed;
    char *cend;
} pgoblin_style;

typedef struct pgoblin_conn {
    u_int32_t        flags;
#   define           PGOBLIN_INT64TIME    0x000010 /* DB known to have int64 timestamp */
#   define           PGOBLIN_FLOATTIME    0x000020 /* DB known to have float timestamp */
    char            *dbname;   /* For libpq         */
    char            *host;     /* For libpq         */
    char            *port;     /* For libpq         */
    char            *username; /* For libpq         */
    PGconn          *conn;
    u_char           intran;   /* Transaction level */
} pgoblin_conn;

typedef struct pgoblin_con {
    u_int32_t        flags;
#   define           PGOBLIN_DUMP_PARSED  0x000001 /* Dump parsed program                 */
#   define           PGOBLIN_COPEND_PASS  0x000002 /* Pass through unconsumed end of file */
#   define           PGOBLIN_COPEND_APND  0x000004 /* Append EOF marker to COPY out file  */
#   define           PGOBLIN_OLDSTYLE     0x800000 /* Redirect execed output to 0 reg_in  */
    u_char           vers[4];                      /* Version           */
    u_char           delim[4];                     /* Delimiter         */
    pgoblin_p       *pgm;
    ssize_t          ppoint;
    size_t           sp;
    char           **stack;
    u_int32_t        maxtmout;
    pgoblin_io       io   [PGOBLIN_REGSIZE];
    pgoblin_conn     conn [PGOBLIN_REGSIZE];
    pgoblin_style   *style[PGOBLIN_REGSIZE];       /* Tables style      */
    char           **argv;
    char           **envp;
} pgoblin_con;

typedef struct pgoblin_c {
    u_int32_t flags;
#   define    PGOBLIN_STYLEN   0x000003
#   define    PGOBLIN_ISSTYLE  0x000004
#   define    PGOBLIN_ISOUT    0x000008
#   define    PGOBLIN_ISCTL    0x000010
#   define    PGOBLIN_ISIN     0x000020
#   define    PGOBLIN_ISCONN   0x000040
#   define    PGOBLIN_IGNORE   0x020000
#   define    PGOBLIN_ILLEGAL  0x040000
#   define    PGOBLIN_NOLITER  0x080000
    u_char    ch[4];
    int     (*car)(pgoblin_con *options, int a[]);
    char     *str;
} pgoblin_c;

pgoblin_p   *pgoblin_init __P((pgoblin_con **options, u_int32_t flags));
int        pgoblin_parser __P((pgoblin_con *options, pgoblin_p *pgm, mife_descriptor *in));
void        pgoblin_Vdump __P((pgoblin_con *options, pgoblin_p *pgm));
int       pgoblin_execute __P((pgoblin_con *options));
